# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest
import numpy

import awkward1

def test_ByteMaskedArray():
    content = awkward1.from_iter([[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False)
    mask = awkward1.layout.Index8(numpy.array([0, 0, 1, 1, 0], dtype=numpy.int8))
    array = awkward1.layout.ByteMaskedArray(mask, content, valid_when=False)
    assert awkward1.to_list(array) == [[0.0, 1.1, 2.2], [], None, None, [6.6, 7.7, 8.8, 9.9]]
    assert awkward1.to_list(array[-1]) == [6.6, 7.7, 8.8, 9.9]
    assert awkward1.to_list(array[-2]) == None
    assert awkward1.to_list(array[1:]) == [[], None, None, [6.6, 7.7, 8.8, 9.9]]
    assert array[4, 1] == 7.7
    assert awkward1.to_list(array[2:, 1]) == [None, None, 7.7]
    assert awkward1.to_list(array[2:, [2, 1, 1, 0]]) == [None, None, [8.8, 7.7, 7.7, 6.6]]

    content = awkward1.from_iter([{"x": 0.0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": [2, 2]}, {"x": 3.3, "y": [3, 3, 3]}, {"x": 4.4, "y": [4, 4, 4, 4]}], highlevel=False)
    mask = awkward1.layout.Index8(numpy.array([0, 0, 1, 1, 0], dtype=numpy.int8))
    array = awkward1.layout.ByteMaskedArray(mask, content, valid_when=False)
    assert awkward1.to_list(array) == [{"x": 0.0, "y": []}, {"x": 1.1, "y": [1]}, None, None, {"x": 4.4, "y": [4, 4, 4, 4]}]
    assert awkward1.to_list(array["x"]) == [0.0, 1.1, None, None, 4.4]
    assert awkward1.to_list(array[["x", "y"]]) == [{"x": 0.0, "y": []}, {"x": 1.1, "y": [1]}, None, None, {"x": 4.4, "y": [4, 4, 4, 4]}]

def test_ByteMaskedArray_jaggedslice0():
    array = awkward1.from_iter([[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False)
    index = awkward1.layout.Index64(numpy.array([0, 1, 2, 3], dtype=numpy.int64))
    indexedarray = awkward1.layout.IndexedOptionArray64(index, array)
    assert awkward1.to_list(indexedarray) == [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]]
    assert awkward1.to_list(indexedarray[awkward1.Array([[0, -1], [0], [], [1, 1]])]) == [[0.0, 2.2], [3.3], [], [7.7, 7.7]]

    mask = awkward1.layout.Index8(numpy.array([0, 0, 0, 0], dtype=numpy.int8))
    maskedarray = awkward1.layout.ByteMaskedArray(mask, array, valid_when=False)
    assert awkward1.to_list(maskedarray) == [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]]
    assert awkward1.to_list(maskedarray[awkward1.Array([[0, -1], [0], [], [1, 1]])]) == [[0.0, 2.2], [3.3], [], [7.7, 7.7]]

def test_ByteMaskedArray_jaggedslice1():
    model = awkward1.Array([[0.0, 1.1, None, 2.2], [], [3.3, None, 4.4], [5.5], [6.6, 7.7, None, 8.8, 9.9]])
    assert awkward1.to_list(model[awkward1.Array([[3, 2, 1, 1, 0], [], [1], [0, 0], [1, 2]])]) == [[2.2, None, 1.1, 1.1, 0.0], [], [None], [5.5, 5.5], [7.7, None]]

    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 999, 2.2, 3.3, 123, 4.4, 5.5, 6.6, 7.7, 321, 8.8, 9.9]))
    mask = awkward1.layout.Index8(numpy.array([0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0], dtype=numpy.int8))
    maskedarray = awkward1.layout.ByteMaskedArray(mask, content, valid_when=False)
    offsets = awkward1.layout.Index64(numpy.array([0, 4, 4, 7, 8, 13], dtype=numpy.int64))
    listarray = awkward1.Array(awkward1.layout.ListOffsetArray64(offsets, maskedarray))
    assert awkward1.to_list(listarray) == awkward1.to_list(model)
    assert awkward1.to_list(listarray[awkward1.Array([[3, 2, 1, 1, 0], [], [1], [0, 0], [1, 2]])]) == [[2.2, None, 1.1, 1.1, 0.0], [], [None], [5.5, 5.5], [7.7, None]]

def test_ByteMaskedArray_jaggedslice2():
    model = awkward1.Array([[[0.0, 1.1, None, 2.2], [], [3.3, None, 4.4]], [], [[5.5]], [[6.6, 7.7, None, 8.8, 9.9]]])
    assert awkward1.to_list(model[awkward1.Array([[[3, 2, 1, 1, 0], [], [1]], [], [[0, 0]], [[1, 2]]])]) == [[[2.2, None, 1.1, 1.1, 0.0], [], [None]], [], [[5.5, 5.5]], [[7.7, None]]]

    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 999, 2.2, 3.3, 123, 4.4, 5.5, 6.6, 7.7, 321, 8.8, 9.9]))
    mask = awkward1.layout.Index8(numpy.array([0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0], dtype=numpy.int8))
    maskedarray = awkward1.layout.ByteMaskedArray(mask, content, valid_when=False)
    offsets = awkward1.layout.Index64(numpy.array([0, 4, 4, 7, 8, 13], dtype=numpy.int64))
    sublistarray = awkward1.layout.ListOffsetArray64(offsets, maskedarray)
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 3, 4, 5], dtype=numpy.int64))
    listarray = awkward1.Array(awkward1.layout.ListOffsetArray64(offsets2, sublistarray))
    assert awkward1.to_list(listarray) == awkward1.to_list(model)
    assert awkward1.to_list(listarray[awkward1.Array([[[3, 2, 1, 1, 0], [], [1]], [], [[0, 0]], [[1, 2]]])]) == [[[2.2, None, 1.1, 1.1, 0.0], [], [None]], [], [[5.5, 5.5]], [[7.7, None]]]

def test_ByteMaskedArray_jaggedslice3():
    model = awkward1.Array([[[[0.0, 1.1, None, 2.2], [], [3.3, None, 4.4]], []], [[[5.5]], [[6.6, 7.7, None, 8.8, 9.9]]]])
    assert awkward1.to_list(model[awkward1.Array([[[[3, 2, 1, 1, 0], [], [1]], []], [[[0, 0]], [[1, 2]]]])]) == [[[[2.2, None, 1.1, 1.1, 0.0], [], [None]], []], [[[5.5, 5.5]], [[7.7, None]]]]

    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 999, 2.2, 3.3, 123, 4.4, 5.5, 6.6, 7.7, 321, 8.8, 9.9]))
    mask = awkward1.layout.Index8(numpy.array([0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0], dtype=numpy.int8))
    maskedarray = awkward1.layout.ByteMaskedArray(mask, content, valid_when=False)
    offsets = awkward1.layout.Index64(numpy.array([0, 4, 4, 7, 8, 13], dtype=numpy.int64))
    subsublistarray = awkward1.layout.ListOffsetArray64(offsets, maskedarray)
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 3, 4, 5], dtype=numpy.int64))
    sublistarray = awkward1.layout.ListOffsetArray64(offsets2, subsublistarray)
    offsets3 = awkward1.layout.Index64(numpy.array([0, 2, 4], dtype=numpy.int64))
    listarray = awkward1.Array(awkward1.layout.ListOffsetArray64(offsets3, sublistarray))
    assert awkward1.to_list(listarray) == awkward1.to_list(model)
    assert awkward1.to_list(listarray[awkward1.Array([[[[3, 2, 1, 1, 0], [], [1]], []], [[[0, 0]], [[1, 2]]]])]) == [[[[2.2, None, 1.1, 1.1, 0.0], [], [None]], []], [[[5.5, 5.5]], [[7.7, None]]]]

def test_ByteMaskedArray_to_slice():
    content = awkward1.layout.NumpyArray(numpy.array([5, 2, 999, 3, 9, 123, 1], dtype=numpy.int64))
    mask = awkward1.layout.Index8(numpy.array([0, 0, 1, 0, 0, 1, 0], dtype=numpy.int8))
    maskedarray = awkward1.layout.ByteMaskedArray(mask, content, valid_when=False)
    assert awkward1.to_list(maskedarray) == [5, 2, None, 3, 9, None, 1]

    assert awkward1._ext._slice_tostring(maskedarray) == "[missing([0, 1, -1, ..., 3, -1, 4], array([5, 2, 3, 9, 1]))]"

def test_ByteMaskedArray_as_slice():
    array = awkward1.Array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], check_valid=True)
    slicecontent = awkward1.from_iter([3, 6, 999, 123, -2, 6], highlevel=False)
    slicemask = awkward1.layout.Index8(numpy.array([0, 0, 1, 1, 0, 0], dtype=numpy.int8))
    slicearray = awkward1.layout.ByteMaskedArray(slicemask, slicecontent, valid_when=False)

    assert awkward1.to_list(array[slicearray]) == [3.3, 6.6, None, None, 8.8, 6.6]

def test_ByteMaskedArray_setidentities():
    content = awkward1.from_iter([[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9], [321]], highlevel=False)
    mask = awkward1.layout.Index8(numpy.array([0, 0, 1, 1, 0], dtype=numpy.int8))
    array = awkward1.layout.ByteMaskedArray(mask, content, valid_when=False)
    assert awkward1.to_list(array) == [[0.0, 1.1, 2.2], [], None, None, [6.6, 7.7, 8.8, 9.9]]
    array.setidentities()
    assert numpy.asarray(array.identities).tolist() == [[0], [1], [2], [3], [4]]
    assert numpy.asarray(array.content.identities).tolist() ==[[0], [1], [2], [3], [4], [-1]]

    assert awkward1.is_valid(awkward1.Array(array))

def test_ByteMaskedArray_num():
    content = awkward1.from_iter([[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], [[5.5]], [[6.6, 7.7, 8.8, 9.9]], [[], [10.0, 11.1, 12.2]]], highlevel=False)
    mask = awkward1.layout.Index8(numpy.array([0, 0, 1, 1, 0], dtype=numpy.int8))
    array = awkward1.Array(awkward1.layout.ByteMaskedArray(mask, content, valid_when=False))
    assert awkward1.to_list(array) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None, None, [[], [10.0, 11.1, 12.2]]]
    assert awkward1.num(array, axis=0) == 5
    assert awkward1.num(array, axis=-3) == 5
    assert awkward1.to_list(awkward1.num(array, axis=1)) == [3, 0, None, None, 2]
    assert awkward1.to_list(awkward1.num(array, axis=-2)) == [3, 0, None, None, 2]
    assert awkward1.to_list(awkward1.num(array, axis=2)) == [[3, 0, 2], [], None, None, [0, 3]]
    assert awkward1.to_list(awkward1.num(array, axis=-1)) == [[3, 0, 2], [], None, None, [0, 3]]

def test_ByteMaskedArray_flatten():
    content = awkward1.from_iter([[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], [[5.5]], [[6.6, 7.7, 8.8, 9.9]], [[], [10.0, 11.1, 12.2]]], highlevel=False)
    mask = awkward1.layout.Index8(numpy.array([0, 0, 1, 1, 0], dtype=numpy.int8))
    array = awkward1.Array(awkward1.layout.ByteMaskedArray(mask, content, valid_when=False))
    assert awkward1.to_list(array) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None, None, [[], [10.0, 11.1, 12.2]]]
    assert awkward1.to_list(awkward1.flatten(array, axis=1)) == [[0.0, 1.1, 2.2], [], [3.3, 4.4], [], [10.0, 11.1, 12.2]]
    assert awkward1.to_list(awkward1.flatten(array, axis=-2)) == [[0.0, 1.1, 2.2], [], [3.3, 4.4], [], [10.0, 11.1, 12.2]]
    assert awkward1.to_list(awkward1.flatten(array, axis=2)) == [[0.0, 1.1, 2.2, 3.3, 4.4], [], None, None, [10.0, 11.1, 12.2]]
    assert awkward1.to_list(awkward1.flatten(array, axis=-1)) == [[0.0, 1.1, 2.2, 3.3, 4.4], [], None, None, [10.0, 11.1, 12.2]]

def test_IndexedOptionArray_pad_none():
    array = awkward1.Array([[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None, None, [[], [10.0, 11.1, 12.2]]])
    assert awkward1.to_list(array) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None, None, [[], [10.0, 11.1, 12.2]]]
    assert awkward1.to_list(awkward1.pad_none(array, 7, axis=0)) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None, None, [[], [10.0, 11.1, 12.2]], None, None]
    assert awkward1.to_list(awkward1.pad_none(array, 7, axis=-3)) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None, None, [[], [10.0, 11.1, 12.2]], None, None]
    assert awkward1.to_list(awkward1.pad_none(array, 3, axis=1)) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [None, None, None], None, None, [[], [10.0, 11.1, 12.2], None]]
    assert awkward1.to_list(awkward1.pad_none(array, 3, axis=-2)) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [None, None, None], None, None, [[], [10.0, 11.1, 12.2], None]]
    assert awkward1.to_list(awkward1.pad_none(array, 3, axis=2)) == [[[0.0, 1.1, 2.2], [None, None, None], [3.3, 4.4, None]], [], None, None, [[None, None, None], [10.0, 11.1, 12.2]]]
    assert awkward1.to_list(awkward1.pad_none(array, 3, axis=-1)) == [[[0.0, 1.1, 2.2], [None, None, None], [3.3, 4.4, None]], [], None, None, [[None, None, None], [10.0, 11.1, 12.2]]]
    assert awkward1.to_list(awkward1.pad_none(array, 3, axis=0, clip=True)) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None]
    assert awkward1.to_list(awkward1.pad_none(array, 3, axis=-3, clip=True)) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None]
    assert awkward1.to_list(awkward1.pad_none(array, 2, axis=1, clip=True)) == [[[0.0, 1.1, 2.2], []], [None, None], None, None, [[], [10.0, 11.1, 12.2]]]
    assert awkward1.to_list(awkward1.pad_none(array, 2, axis=-2, clip=True)) == [[[0.0, 1.1, 2.2], []], [None, None], None, None, [[], [10.0, 11.1, 12.2]]]
    assert awkward1.to_list(awkward1.pad_none(array, 2, axis=2, clip=True)) == [[[0.0, 1.1], [None, None], [3.3, 4.4]], [], None, None, [[None, None], [10.0, 11.1]]]
    assert awkward1.to_list(awkward1.pad_none(array, 2, axis=-1, clip=True)) == [[[0.0, 1.1], [None, None], [3.3, 4.4]], [], None, None, [[None, None], [10.0, 11.1]]]

def test_ByteMaskedArray_pad_none():
    content = awkward1.from_iter([[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], [[5.5]], [[6.6, 7.7, 8.8, 9.9]], [[], [10.0, 11.1, 12.2]]], highlevel=False)
    mask = awkward1.layout.Index8(numpy.array([0, 0, 1, 1, 0], dtype=numpy.int8))
    array = awkward1.Array(awkward1.layout.ByteMaskedArray(mask, content, valid_when=False))
    assert awkward1.to_list(array) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None, None, [[], [10.0, 11.1, 12.2]]]
    assert awkward1.to_list(awkward1.pad_none(array, 7, axis=0)) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None, None, [[], [10.0, 11.1, 12.2]], None, None]
    assert awkward1.to_list(awkward1.pad_none(array, 7, axis=-3)) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None, None, [[], [10.0, 11.1, 12.2]], None, None]
    assert awkward1.to_list(awkward1.pad_none(array, 3, axis=1)) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [None, None, None], None, None, [[], [10.0, 11.1, 12.2], None]]
    assert awkward1.to_list(awkward1.pad_none(array, 3, axis=-2)) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [None, None, None], None, None, [[], [10.0, 11.1, 12.2], None]]
    assert awkward1.to_list(awkward1.pad_none(array, 3, axis=2)) == [[[0.0, 1.1, 2.2], [None, None, None], [3.3, 4.4, None]], [], None, None, [[None, None, None], [10.0, 11.1, 12.2]]]
    assert awkward1.to_list(awkward1.pad_none(array, 3, axis=-1)) == [[[0.0, 1.1, 2.2], [None, None, None], [3.3, 4.4, None]], [], None, None, [[None, None, None], [10.0, 11.1, 12.2]]]
    assert awkward1.to_list(awkward1.pad_none(array, 3, axis=0, clip=True)) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None]
    assert awkward1.to_list(awkward1.pad_none(array, 3, axis=-3, clip=True)) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None]
    assert awkward1.to_list(awkward1.pad_none(array, 2, axis=1, clip=True)) == [[[0.0, 1.1, 2.2], []], [None, None], None, None, [[], [10.0, 11.1, 12.2]]]
    assert awkward1.to_list(awkward1.pad_none(array, 2, axis=-2, clip=True)) == [[[0.0, 1.1, 2.2], []], [None, None], None, None, [[], [10.0, 11.1, 12.2]]]
    assert awkward1.to_list(awkward1.pad_none(array, 2, axis=2, clip=True)) == [[[0.0, 1.1], [None, None], [3.3, 4.4]], [], None, None, [[None, None], [10.0, 11.1]]]
    assert awkward1.to_list(awkward1.pad_none(array, 2, axis=-1, clip=True)) == [[[0.0, 1.1], [None, None], [3.3, 4.4]], [], None, None, [[None, None], [10.0, 11.1]]]

def test_ByteMaskedArray_reduce():
    content = awkward1.layout.NumpyArray(numpy.array([2, 3, 5, 7, 11, 0, 0, 0, 0, 0, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 0, 0, 0, 0, 0, 101, 103, 107, 109, 113], dtype=numpy.int64))
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 10, 15, 20, 25, 30], dtype=numpy.int64))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets1, content)
    # index = awkward1.layout.Index64(numpy.array([0, -1, 2, 3, -1, 5], dtype=numpy.int64))
    # optionarray = awkward1.layout.IndexedOptionArray64(index, listoffsetarray)
    mask = awkward1.layout.Index8(numpy.array([0, 1, 0, 0, 1, 0], dtype=numpy.int8))
    optionarray = awkward1.layout.ByteMaskedArray(mask, listoffsetarray, valid_when=False)
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, optionarray)
    assert awkward1.to_list(depth2) == [
        [[  2,   3,   5,   7,  11],
         None,
         [ 31,  37,  41,  43,  47]],
        [[ 53,  59,  61,  67,  71],
         None,
         [101, 103, 107, 109, 113]]]

    assert awkward1.to_list(depth2.prod(-1)) == [
        [  2 *   3 *   5 *   7 *  11,
         None,
          31 *  37 *  41 *  43 *  47],
        [ 53 *  59 *  61 *  67 *  71,
         None,
         101 * 103 * 107 * 109 * 113]]

    assert awkward1.to_list(depth2.prod(-2)) == [
        [  31*2,   37*3,   41*5,   43*7,  47*11],
        [101*53, 103*59, 107*61, 109*67, 113*71]]

    assert awkward1.to_list(depth2.prod(-3)) == [
        [  53*2,   59*3,   61*5,   67*7,  71*11],
        [],
        [101*31, 103*37, 107*41, 109*43, 113*47]]

    content = awkward1.layout.NumpyArray(numpy.array([2, 3, 5, 7, 11, 0, 0, 0, 0, 0, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 0, 0, 0, 0, 0, 101, 103, 107, 109, 113], dtype=numpy.int64))
    # index = awkward1.layout.Index64(numpy.array([0, 1, 2, 3, 4, -1, -1, -1, -1, -1, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, -1, -1, -1, -1, -1, 25, 26, 27, 28, 29], dtype=numpy.int64))
    # optionarray = awkward1.layout.IndexedOptionArray64(index, content)
    mask = awkward1.layout.Index8(numpy.array([0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0], dtype=numpy.int8))
    optionarray = awkward1.layout.ByteMaskedArray(mask, content, valid_when=False)
    offsets1 = awkward1.layout.Index64(numpy.array([0, 5, 10, 15, 20, 25, 30], dtype=numpy.int64))
    listoffsetarray = awkward1.layout.ListOffsetArray64(offsets1, optionarray)
    offsets2 = awkward1.layout.Index64(numpy.array([0, 3, 6], dtype=numpy.int64))
    depth2 = awkward1.layout.ListOffsetArray64(offsets2, listoffsetarray)
    assert awkward1.to_list(depth2) == [
        [[   2,    3,    5,    7,   11],
         [None, None, None, None, None],
         [  31,   37,   41,   43,   47]],
        [[  53,   59,   61,   67,   71],
         [None, None, None, None, None],
         [ 101,  103,  107,  109,  113]]]

    assert awkward1.to_list(depth2.prod(-1)) == [
        [  2 *   3 *   5 *   7 *  11,
           1 *   1 *   1 *   1 *   1,
          31 *  37 *  41 *  43 *  47],
        [ 53 *  59 *  61 *  67 *  71,
           1 *   1 *   1 *   1 *   1,
         101 * 103 * 107 * 109 * 113]]

    assert awkward1.to_list(depth2.prod(-2)) == [
        [  31*2,   37*3,   41*5,   43*7,  47*11],
        [101*53, 103*59, 107*61, 109*67, 113*71]]

    assert awkward1.to_list(depth2.prod(-3)) == [
        [  53*2,   59*3,   61*5,   67*7,  71*11],
        [     1,      1,      1,      1,      1],
        [101*31, 103*37, 107*41, 109*43, 113*47]]

def test_ByteMaskedArray_localindex():
    content = awkward1.from_iter([[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], [[5.5]], [[6.6, 7.7, 8.8, 9.9]], [[], [10.0, 11.1, 12.2]]], highlevel=False)
    mask = awkward1.layout.Index8(numpy.array([0, 0, 1, 1, 0], dtype=numpy.int8))
    array = awkward1.layout.ByteMaskedArray(mask, content, valid_when=False)
    assert awkward1.to_list(array) == [[[0.0, 1.1, 2.2], [], [3.3, 4.4]], [], None, None, [[], [10.0, 11.1, 12.2]]]
    assert awkward1.to_list(array.localindex(axis=0)) == [0, 1, 2, 3, 4]
    assert awkward1.to_list(array.localindex(axis=-3)) == [0, 1, 2, 3, 4]
    assert awkward1.to_list(array.localindex(axis=1)) == [[0, 1, 2], [], None, None, [0, 1]]
    assert awkward1.to_list(array.localindex(axis=-2)) == [[0, 1, 2], [], None, None, [0, 1]]
    assert awkward1.to_list(array.localindex(axis=2)) == [[[0, 1, 2], [], [0, 1]], [], None, None, [[], [0, 1, 2]]]
    assert awkward1.to_list(array.localindex(axis=-1)) == [[[0, 1, 2], [], [0, 1]], [], None, None, [[], [0, 1, 2]]]

def test_ByteMaskedArray_combinations():
    content = awkward1.from_iter([[[0, 1, 2], [], [3, 4]], [], [[5]], [[6, 7, 8, 9]], [[], [10, 11, 12]]], highlevel=False)
    mask = awkward1.layout.Index8(numpy.array([0, 0, 1, 1, 0], dtype=numpy.int8))
    array = awkward1.Array(awkward1.layout.ByteMaskedArray(mask, content, valid_when=False))
    assert awkward1.to_list(array) == [[[0, 1, 2], [], [3, 4]], [], None, None, [[], [10, 11, 12]]]
    assert awkward1.to_list(awkward1.combinations(array, 2, axis=0)) == [([[0, 1, 2], [], [3, 4]], []), ([[0, 1, 2], [], [3, 4]], None), ([[0, 1, 2], [], [3, 4]], None), ([[0, 1, 2], [], [3, 4]], [[], [10, 11, 12]]), ([], None), ([], None), ([], [[], [10, 11, 12]]), (None, None), (None, [[], [10, 11, 12]]), (None, [[], [10, 11, 12]])]
    assert awkward1.to_list(awkward1.combinations(array, 2, axis=-3)) == [([[0, 1, 2], [], [3, 4]], []), ([[0, 1, 2], [], [3, 4]], None), ([[0, 1, 2], [], [3, 4]], None), ([[0, 1, 2], [], [3, 4]], [[], [10, 11, 12]]), ([], None), ([], None), ([], [[], [10, 11, 12]]), (None, None), (None, [[], [10, 11, 12]]), (None, [[], [10, 11, 12]])]
    assert awkward1.to_list(awkward1.combinations(array, 2, axis=1)) == [[([0, 1, 2], []), ([0, 1, 2], [3, 4]), ([], [3, 4])], [], None, None, [([], [10, 11, 12])]]
    assert awkward1.to_list(awkward1.combinations(array, 2, axis=-2)) == [[([0, 1, 2], []), ([0, 1, 2], [3, 4]), ([], [3, 4])], [], None, None, [([], [10, 11, 12])]]
    assert awkward1.to_list(awkward1.combinations(array, 2, axis=2)) == [[[(0, 1), (0, 2), (1, 2)], [], [(3, 4)]], [], None, None, [[], [(10, 11), (10, 12), (11, 12)]]]
    assert awkward1.to_list(awkward1.combinations(array, 2, axis=-1)) == [[[(0, 1), (0, 2), (1, 2)], [], [(3, 4)]], [], None, None, [[], [(10, 11), (10, 12), (11, 12)]]]

def test_IndexedOptionArray_combinations():
    content = awkward1.from_iter([[[0, 1, 2], [], [3, 4]], [], [[5]], [[6, 7, 8, 9]], [[], [10, 11, 12]]], highlevel=False)
    index = awkward1.layout.Index64(numpy.array([0, 1, -1, -1, 4], dtype=numpy.int64))
    array = awkward1.Array(awkward1.layout.IndexedOptionArray64(index, content))
    assert awkward1.to_list(array) == [[[0, 1, 2], [], [3, 4]], [], None, None, [[], [10, 11, 12]]]
    assert awkward1.to_list(awkward1.combinations(array, 2, axis=0)) == [([[0, 1, 2], [], [3, 4]], []), ([[0, 1, 2], [], [3, 4]], None), ([[0, 1, 2], [], [3, 4]], None), ([[0, 1, 2], [], [3, 4]], [[], [10, 11, 12]]), ([], None), ([], None), ([], [[], [10, 11, 12]]), (None, None), (None, [[], [10, 11, 12]]), (None, [[], [10, 11, 12]])]
    assert awkward1.to_list(awkward1.combinations(array, 2, axis=-3)) == [([[0, 1, 2], [], [3, 4]], []), ([[0, 1, 2], [], [3, 4]], None), ([[0, 1, 2], [], [3, 4]], None), ([[0, 1, 2], [], [3, 4]], [[], [10, 11, 12]]), ([], None), ([], None), ([], [[], [10, 11, 12]]), (None, None), (None, [[], [10, 11, 12]]), (None, [[], [10, 11, 12]])]
    assert awkward1.to_list(awkward1.combinations(array, 2, axis=1)) == [[([0, 1, 2], []), ([0, 1, 2], [3, 4]), ([], [3, 4])], [], None, None, [([], [10, 11, 12])]]
    assert awkward1.to_list(awkward1.combinations(array, 2, axis=-2)) == [[([0, 1, 2], []), ([0, 1, 2], [3, 4]), ([], [3, 4])], [], None, None, [([], [10, 11, 12])]]
    assert awkward1.to_list(awkward1.combinations(array, 2, axis=2)) == [[[(0, 1), (0, 2), (1, 2)], [], [(3, 4)]], [], None, None, [[], [(10, 11), (10, 12), (11, 12)]]]
    assert awkward1.to_list(awkward1.combinations(array, 2, axis=-1)) == [[[(0, 1), (0, 2), (1, 2)], [], [(3, 4)]], [], None, None, [[], [(10, 11), (10, 12), (11, 12)]]]

def test_merge():
    content = awkward1.from_iter([[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False)
    mask = awkward1.layout.Index8(numpy.array([0, 0, 1, 1, 0], dtype=numpy.int8))
    array1 = awkward1.layout.ByteMaskedArray(mask, content, valid_when=False)
    assert awkward1.to_list(array1) == [[0.0, 1.1, 2.2], [], None, None, [6.6, 7.7, 8.8, 9.9]]
    array2 = awkward1.Array([[0.0, 1.1, 2.2], [], None, None, [6.6, 7.7, 8.8, 9.9]])
    array12 = awkward1.concatenate([array1, array2], highlevel=False)
    assert awkward1.to_list(array12) == [[0.0, 1.1, 2.2], [], None, None, [6.6, 7.7, 8.8, 9.9], [0.0, 1.1, 2.2], [], None, None, [6.6, 7.7, 8.8, 9.9]]
    assert isinstance(array12, awkward1.layout.IndexedOptionArray64)
    assert isinstance(array12.content, (awkward1.layout.ListArray64, awkward1.layout.ListOffsetArray64))
    assert isinstance(array12.content.content, awkward1.layout.NumpyArray)
    assert awkward1.to_list(array12.content.content) == [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 0.0, 1.1, 2.2, 6.6, 7.7, 8.8, 9.9]

def test_BitMaskedArray():
    content = awkward1.layout.NumpyArray(numpy.arange(13))
    mask = awkward1.layout.IndexU8(numpy.array([58, 59], dtype=numpy.uint8))
    array = awkward1.layout.BitMaskedArray(mask, content, valid_when=False, length=13, lsb_order=False)
    assert numpy.asarray(array.bytemask()).tolist() == [0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 1]
    assert numpy.asarray(array.toByteMaskedArray().mask).tolist() == [0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 1, 1, 1]
    assert numpy.asarray(array.toIndexedOptionArray64().index).tolist() == [0, 1, -1, -1, -1, 5, -1, 7, 8, 9, -1, -1, -1]
    assert awkward1.to_list(array) == [0, 1, None, None, None, 5, None, 7, 8, 9, None, None, None]
    assert awkward1.to_json(array) == "[0,1,null,null,null,5,null,7,8,9,null,null,null]"
    assert awkward1.to_list(array[1:-1]) == [1, None, None, None, 5, None, 7, 8, 9, None, None]
    assert awkward1.to_list(array[8:]) == [8, 9, None, None, None]

    array = awkward1.layout.BitMaskedArray(mask, content, valid_when=False, length=13, lsb_order=True)
    assert numpy.asarray(array.bytemask()).tolist() == [0, 1, 0, 1, 1, 1, 0, 0, 1, 1, 0, 1, 1]
    assert numpy.asarray(array.toByteMaskedArray().mask).tolist() == [0, 1, 0, 1, 1, 1, 0, 0, 1, 1, 0, 1, 1]
    assert numpy.asarray(array.toIndexedOptionArray64().index).tolist() == [0, -1, 2, -1, -1, -1, 6, 7, -1, -1, 10, -1, -1]
    assert awkward1.to_list(array) == [0, None, 2, None, None, None, 6, 7, None, None, 10, None, None]
    assert awkward1.to_json(array) == "[0,null,2,null,null,null,6,7,null,null,10,null,null]"
    assert awkward1.to_list(array[1:-1]) == [None, 2, None, None, None, 6, 7, None, None, 10, None]
    assert awkward1.to_list(array[8:]) == [None, None, 10, None, None]

def test_UnmaskedArray():
    content = awkward1.layout.NumpyArray(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5], dtype=numpy.float64))
    array = awkward1.layout.UnmaskedArray(content)
    assert awkward1.to_list(array) == [1.1, 2.2, 3.3, 4.4, 5.5]
    assert str(awkward1.type(content)) == "float64"
    assert str(awkward1.type(awkward1.Array(content))) == "5 * float64"
    assert str(awkward1.type(array)) == "?float64"
    assert str(awkward1.type(awkward1.Array(array))) == "5 * ?float64"

def test_tomask():
    array = awkward1.Array([[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]])
    mask1 = awkward1.Array([True, True, False, False, True])
    assert awkward1.to_list(array[mask1]) == [[0.0, 1.1, 2.2], [], [6.6, 7.7, 8.8, 9.9]]
    assert awkward1.to_list(awkward1.mask(array, mask1)) == [[0.0, 1.1, 2.2], [], None, None, [6.6, 7.7, 8.8, 9.9]]

    mask2 = awkward1.Array([[False, True, False], [], [True, True], [False], [True, False, False, True]])
    assert awkward1.to_list(array[mask2]) == [[1.1], [], [3.3, 4.4], [], [6.6, 9.9]]
    assert awkward1.to_list(awkward1.mask(array, mask2)) == [[None, 1.1, None], [], [3.3, 4.4], [None], [6.6, None, None, 9.9]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest
import numpy

import awkward1

def test():
    empty1 = awkward1.Array(awkward1.layout.EmptyArray(), check_valid=True)
    empty2 = awkward1.Array(awkward1.layout.ListOffsetArray64(awkward1.layout.Index64(numpy.array([0, 0, 0, 0], dtype=numpy.int64)), awkward1.layout.EmptyArray()), check_valid=True)
    array = awkward1.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]], check_valid=True)

    awkward1.to_numpy(empty1).dtype.type is numpy.float64

    awkward1.to_list(array[empty1]) == []
    awkward1.to_list(array[empty1,]) == []
    awkward1.to_list(array[empty2]) == [[], [], []]
    awkward1.to_list(array[empty2,]) == [[], [], []]



# BSD 3-Clause License; see https://github.com/jpivarski/awkward-1.0/blob/master/LICENSE

import glob
import os
import platform
import subprocess
import sys
import distutils.util
import multiprocessing
import shutil

import setuptools
import setuptools.command.build_ext
import setuptools.command.install

from setuptools import setup, Extension

install_requires = open("requirements.txt").read().strip().split()

extras = {"test": open("requirements-test.txt").read().strip().split(),
          "dev":  ['numba>=0.50.0;python_version>="3.6"',
                   'pandas>=0.24.0;python_version>="3.6"',
                   'numexpr;python_version>="3.6"',
                   'autograd;python_version>="3.6"',
                   'pyarrow>=1.0;python_version>="3.6" and sys_platform != "win32"']}
extras["all"] = sum(extras.values(), [])

tests_require = extras["test"]


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=""):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)

class CMakeBuild(setuptools.command.build_ext.build_ext):
    def build_extensions(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " + ", ".join(x.name for x in self.extensions))

        for x in self.extensions:
            self.build_extension(x)

    def build_extension(self, ext):

        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ["-DCMAKE_INSTALL_PREFIX={0}".format(extdir),
                      "-DPYTHON_EXECUTABLE={0}".format(sys.executable),
                      "-DPYBUILD=ON",
                      "-DCMAKE_OSX_DEPLOYMENT_TARGET=10.9",
                      "-DPYBUILD=ON",
                      "-DBUILD_TESTING=OFF"]
        try:
           compiler_path = self.compiler.compiler_cxx[0]
           cmake_args.append("-DCMAKE_CXX_COMPILER={0}".format(compiler_path))
        except AttributeError:
            print("Not able to access compiler path (on Windows), using CMake default")

        cfg = "Debug" if self.debug else "Release"
        build_args = ["--config", cfg]

        if platform.system() == "Windows":
            cmake_args += ["-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{0}={1}".format(cfg.upper(), extdir),
                           "-DCMAKE_WINDOWS_EXPORT_ALL_SYMBOLS=TRUE"]
            if sys.maxsize > 2**32:
                cmake_args += ["-A", "x64"]
            build_args += ["--", "/m"]
        else:
            cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]

        if platform.system() == "Windows":
            build_args += ["/m"]
        else:
            build_args += ["-j", str(multiprocessing.cpu_count())]

        if not os.path.exists(self.build_temp):
             os.makedirs(self.build_temp)
        build_dir = self.build_temp

        # for scikit-build:
        # build_dir = os.path.join(DIR, "_pybuild")

        subprocess.check_call(["cmake", "-S", ext.sourcedir, "-B", build_dir] + cmake_args)
        subprocess.check_call(["cmake", "--build", build_dir] + build_args)
        subprocess.check_call(["cmake", "--build", build_dir, "--target", "install"])

if platform.system() == "Windows":
    # Libraries are not installed system-wide on Windows, but they do have to be in the awkward1 directory.
    libraries = []

    def tree(x):
        print(x + (" (dir)" if os.path.isdir(x) else ""))
        if os.path.isdir(x):
            for y in os.listdir(x):
                tree(os.path.join(x, y))

    class Install(setuptools.command.install.install):
        def run(self):
            print("--- build directory -------------------------------------------")
            tree("build")

            print("--- specifically, the dlldir ----------------------------------")
            dlldir = os.path.join(os.path.join("build", "temp.%s-%d.%d" % (distutils.util.get_platform(), sys.version_info[0], sys.version_info[1])), "Release", "Release")
            tree(dlldir)

            print("--- copying ---------------------------------------------------")
            for x in os.listdir(dlldir):
                if x.startswith("awkward"):
                    print("copying", os.path.join(dlldir, x), "-->", os.path.join(self.build_lib, "awkward1", x))
                    shutil.copyfile(os.path.join(dlldir, x), os.path.join(self.build_lib, "awkward1", x))
            print("--- deleting --------------------------------------------------")

            outerdir = os.path.join(os.path.join("build", "lib.%s-%d.%d" % (distutils.util.get_platform(), sys.version_info[0], sys.version_info[1])))
            for x in os.listdir(outerdir):
                if x.endswith(".pyd"):
                    print("deleting", os.path.join(outerdir, x))
                    os.remove(os.path.join(outerdir, x))

            print("--- begin normal install --------------------------------------")
            setuptools.command.install.install.run(self)

else:
    # Libraries do not exist yet, so they cannot be determined with a glob pattern.
    libdir = os.path.join(os.path.join("build", "lib.%s-%d.%d" % (distutils.util.get_platform(), sys.version_info[0], sys.version_info[1])), "awkward1")
    static = ".a"
    if platform.system() == "Darwin":
        shared = ".dylib"
    else:
        shared = ".so"
    libraries = [("lib", [os.path.join(libdir, "libawkward-cpu-kernels-static" + static),
                          os.path.join(libdir, "libawkward-cpu-kernels" + shared),
                          os.path.join(libdir, "libawkward-static" + static),
                          os.path.join(libdir, "libawkward" + shared)])]

    Install = setuptools.command.install.install

setup(name = "awkward1",
      packages = setuptools.find_packages(where="src"),
      package_dir = {"": "src"},
      include_package_data = True,
      package_data = {"": ["*.dll"]},
      data_files = libraries + [
          ("include/awkward",              glob.glob("include/awkward/*.h")),
          ("include/awkward/cpu-kernels",  glob.glob("include/awkward/cpu-kernels/*.h")),
          ("include/awkward/python",       glob.glob("include/awkward/python/*.h")),
          ("include/awkward/array",        glob.glob("include/awkward/array/*.h")),
          ("include/awkward/builder",      glob.glob("include/awkward/builder/*.h")),
          ("include/awkward/partition",    glob.glob("include/awkward/partition/*.h")),
          ("include/awkward/io",           glob.glob("include/awkward/io/*.h")),
          ("include/awkward/type",         glob.glob("include/awkward/type/*.h"))],
      version = open("VERSION_INFO").read().strip(),
      author = "Jim Pivarski",
      author_email = "pivarski@princeton.edu",
      maintainer = "Jim Pivarski",
      maintainer_email = "pivarski@princeton.edu",
      description = "Manipulate JSON-like data with NumPy-like idioms.",
      long_description = """<a href="https://github.com/scikit-hep/awkward-1.0#readme"><img src="https://github.com/scikit-hep/awkward-1.0/raw/master/docs-img/logo/logo-300px.png"></a>

Awkward Array is a library for **nested, variable-sized data**, including arbitrary-length lists, records, mixed types, and missing data, using **NumPy-like idioms**.

Arrays are **dynamically typed**, but operations on them are **compiled and fast**. Their behavior coincides with NumPy when array dimensions are regular and generalizes when they're not.

<table>
  <tr>
    <td width="66%" valign="top">
      <a href="https://awkward-array.org">
        <img src="https://github.com/scikit-hep/awkward-1.0/raw/master/docs-img/panel-tutorials.png" width="570">
      </a>
      <p align="center"><b>
        <a href="https://awkward-array.org">
        How-to tutorials
        </a>
      </b></p>
    </td>
    <td width="33%" valign="top">
      <a href="https://awkward-array.readthedocs.io/en/latest/index.html">
        <img src="https://github.com/scikit-hep/awkward-1.0/raw/master/docs-img/panel-sphinx.png" width="268">
      </a>
      <p align="center"><b>
        <a href="https://awkward-array.readthedocs.io/en/latest/index.html">
        Python API reference
        </a>
      </b></p>
      <a href="https://awkward-array.readthedocs.io/en/latest/_static/index.html">
        <img src="https://github.com/scikit-hep/awkward-1.0/raw/master/docs-img/panel-doxygen.png" width="268">
      </a>
      <p align="center"><b>
        <a href="https://awkward-array.readthedocs.io/en/latest/_static/index.html">
        C++ API reference
        </a>
      </b></p>
    </td>
  </tr>
</table>

# Motivating example

Given an array of objects with `x`, `y` fields and variable-length nested lists like

```python
array = ak.Array([
    [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}, {"x": 3.3, "y": [1, 2, 3]}],
    [],
    [{"x": 4.4, "y": {1, 2, 3, 4]}, {"x": 5.5, "y": [1, 2, 3, 4, 5]}]
])
```

the following slices out the `y` values, drops the first element from each inner list, and runs NumPy's `np.square` function on everything that is left:

```python
output = np.square(array["y", ..., 1:])
```

The result is

```python
[
    [[], [4], [4, 9]],
    [],
    [[4, 9, 16], [4, 9, 16, 25]]
]
```

The equivalent using only Python is

```python
output = []
for sublist in array:
    tmp1 = []
    for record in sublist:
        tmp2 = []
        for number in record["y"][1:]:
            tmp2.append(np.square(number))
        tmp1.append(tmp2)
    output.append(tmp1)
```

Not only is the expression using Awkward Arrays more concise, using idioms familiar from NumPy, but it's much faster and uses less memory.

For a similar problem 10 million times larger than the one above (on a single-threaded 2.2 GHz processor),

   * the Awkward Array one-liner takes **4.6 seconds** to run and uses **2.1 GB** of memory,
   * the equivalent using Python lists and dicts takes **138 seconds** to run and uses **22 GB** of memory.

Speed and memory factors in the double digits are common because we're replacing Python's dynamically typed, pointer-chasing virtual machine with type-specialized, precompiled routines on contiguous data. (In other words, for the same reasons as NumPy.) Even higher speedups are possible when Awkward Array is paired with [Numba](https://numba.pydata.org/).

Our [presentation at SciPy 2020](https://youtu.be/WlnUF3LRBj4) provides a good introduction, showing how to use these arrays in a real analysis.

# Installation

Awkward Array can be installed [from PyPI](https://pypi.org/project/awkward1/) using pip:

```bash
pip install awkward1
```

Most users will get a precompiled binary (wheel) for your operating system and Python version. If not, the above attempts to compile from source.

   * Report bugs, request features, and ask for additional documentation on [GitHub Issues](https://github.com/scikit-hep/awkward-1.0/issues). If you have a general "How do I…?" question, we'll answer it as a new [example in the tutorial](https://awkward-array.org/how-to.html).
   * If you have a problem that's too specific to be new documentation or it isn't exclusively related to Awkward Array, it might be more appropriate to ask on [StackOverflow with the [awkward-array] tag](https://stackoverflow.com/questions/tagged/awkward-array). Be sure to include tags for any other libraries that you use, such as Pandas or PyTorch.
   * The [Gitter Scikit-HEP/community](https://gitter.im/Scikit-HEP/community) is a way to get in touch with all Scikit-HEP developers and users.
""",
      long_description_content_type = "text/markdown",
      url = "https://github.com/scikit-hep/awkward-1.0",
      download_url = "https://github.com/scikit-hep/awkward-1.0/releases",
      license = "BSD 3-clause",
      entry_points = {
        "numba_extensions": ["init = awkward1._connect._numba:register"]
      },
      test_suite = "tests",
      python_requires = ">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
      install_requires = install_requires,
      tests_require = extras["test"],
      extras_require = extras,

      # cmake_args=['-DBUILD_TESTING=OFF'],      # for scikit-build
      ext_modules = [CMakeExtension("awkward")], # NOT scikit-build
      cmdclass = {"build_ext": CMakeBuild,       # NOT scikit-build
                  "install": Install},

      classifiers = [
#         "Development Status :: 1 - Planning",
#         "Development Status :: 2 - Pre-Alpha",
#         "Development Status :: 3 - Alpha",
#         "Development Status :: 4 - Beta",
          "Development Status :: 5 - Production/Stable",
#         "Development Status :: 6 - Mature",
#         "Development Status :: 7 - Inactive",
          "Intended Audience :: Developers",
          "Intended Audience :: Information Technology",
          "Intended Audience :: Science/Research",
          "License :: OSI Approved :: BSD License",
          "Operating System :: MacOS :: MacOS X",
          "Operating System :: Microsoft :: Windows",
          "Operating System :: POSIX :: Linux",
          "Operating System :: Unix",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Information Analysis",
          "Topic :: Scientific/Engineering :: Mathematics",
          "Topic :: Scientific/Engineering :: Physics",
          "Topic :: Software Development",
          "Topic :: Utilities",
          ])



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest
import numpy

import awkward1

numba = pytest.importorskip("numba")

py27 = (sys.version_info[0] < 3)


def test_string():
    array = awkward1.Array(["one", "two", "three", "four", "five"], check_valid=True)

    def f1(x, i):
        return x[i]

    assert f1(array, 0) == "one"
    assert f1(array, 1) == "two"
    assert f1(array, 2) == "three"

    f1 = numba.njit(f1)

    assert f1(array, 0) == "one"
    assert f1(array, 1) == "two"
    assert f1(array, 2) == "three"

    if not py27:
        def f2(x, i, j):
            return x[i] + x[j]

        assert f2(array, 1, 3) == "twofour"
        assert numba.njit(f2)(array, 1, 3) == "twofour"


def test_bytestring():
    array = awkward1.Array([b"one", b"two", b"three", b"four", b"five"], check_valid=True)

    def f1(x, i):
        return x[i]

    assert f1(array, 0) == b"one"
    assert f1(array, 1) == b"two"
    assert f1(array, 2) == b"three"

    f1 = numba.njit(f1)

    assert f1(array, 0) == b"one"
    assert f1(array, 1) == b"two"
    assert f1(array, 2) == b"three"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import distutils.version

# NumPy 1.13.1 introduced NEP13, without which Awkward ufuncs won't work, which
# would be worse than lacking a feature: it would cause unexpected output.
# NumPy 1.17.0 introduced NEP18, which is optional (use ak.* instead of np.*).
import numpy

if distutils.version.LooseVersion(numpy.__version__) < distutils.version.LooseVersion(
    "1.13.1"
):
    raise ImportError("Numpy 1.13.1 or later required")

# NumPy-like alternatives
import awkward1.nplike

# shims for C++ (now everything is compiled into one 'awkward1._ext' module)
import awkward1.layout
import awkward1.types
import awkward1.forms
import awkward1.partition

# third-party connectors
import awkward1._connect._numba

numba = type(awkward1._connect._numba)("numba")
numba.register = awkward1._connect._numba.register

import awkward1._connect._pandas

pandas = type(awkward1._connect._pandas)("pandas")
pandas.register = awkward1._connect._pandas.register
pandas.df = awkward1._connect._pandas.df
pandas.dfs = awkward1._connect._pandas.dfs

import awkward1._connect._numexpr

numexpr = type(awkward1._connect._numexpr)("numexpr")
numexpr.evaluate = awkward1._connect._numexpr.evaluate
numexpr.re_evaluate = awkward1._connect._numexpr.re_evaluate

import awkward1._connect._autograd

autograd = type(awkward1._connect._autograd)("autograd")
autograd.elementwise_grad = awkward1._connect._autograd.elementwise_grad

# high-level interface
behavior = {}
from awkward1.highlevel import Array
from awkward1.highlevel import Record
from awkward1.highlevel import ArrayBuilder

# behaviors
from awkward1.behaviors.mixins import *
from awkward1.behaviors.string import *
from awkward1.behaviors.categorical import *

# operations
from awkward1.operations.convert import *
from awkward1.operations.describe import *
from awkward1.operations.structure import *
from awkward1.operations.reducers import *
from awkward1.operations.transfers import *

# version
__version__ = awkward1._ext.__version__

# call C++ startup function
awkward1._ext.startup()

__all__ = [
    x
    for x in list(globals())
    if not x.startswith("_") and x not in ("distutils", "numpy")
]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward1/blob/master/LICENSE

from __future__ import absolute_import

import os.path
import glob

directory, filename = os.path.split(__file__)

shared_library_path = None
for filename in glob.glob(os.path.join(directory, "*awkward-cuda-kernels.*")):
    shared_library_path = filename

del os
del glob
del directory
del filename




# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/master/LICENSE

from __future__ import absolute_import

import sys

import pytest
import numpy

import awkward1

awkward0 = pytest.importorskip("awkward")

def test_toawkward0():
    array = awkward1.from_iter([1.1, 2.2, 3.3, 4.4], highlevel=False)
    assert isinstance(awkward1.to_awkward0(array), numpy.ndarray)
    assert awkward1.to_awkward0(array).tolist() == [1.1, 2.2, 3.3, 4.4]

    array = awkward1.from_numpy(numpy.arange(2*3*5).reshape(2, 3, 5), highlevel=False).toRegularArray()
    assert isinstance(awkward1.to_awkward0(array), awkward0.JaggedArray)
    assert awkward1.to_awkward0(array).tolist() == [[[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]], [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]]]

    array = awkward1.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    assert isinstance(awkward1.to_awkward0(array), awkward0.JaggedArray)
    assert awkward1.to_awkward0(array).tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]

    array = awkward1.layout.ListArray64(awkward1.layout.Index64(numpy.array([4, 999, 1], dtype=numpy.int64)), awkward1.layout.Index64(numpy.array([7, 999, 3], dtype=numpy.int64)), awkward1.layout.NumpyArray(numpy.array([3.14, 4.4, 5.5, 123, 1.1, 2.2, 3.3, 321])))
    assert isinstance(awkward1.to_awkward0(array), awkward0.JaggedArray)
    assert awkward1.to_awkward0(array).tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]

    array = awkward1.from_iter([{"x": 0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": [2, 2]}, {"x": 3.3, "y": [3, 3, 3]}], highlevel=False)
    assert isinstance(awkward1.to_awkward0(array[2]), dict)
    assert awkward1.to_awkward0(array[2])["x"] == 2.2
    assert isinstance(awkward1.to_awkward0(array[2])["y"], numpy.ndarray)
    assert awkward1.to_awkward0(array[2])["y"].tolist() == [2, 2]

    assert isinstance(awkward1.to_awkward0(array), awkward0.Table)
    assert awkward1.to_awkward0(array).tolist() == [{"x": 0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": [2, 2]}, {"x": 3.3, "y": [3, 3, 3]}]

    array = awkward1.from_iter([(0, []), (1.1, [1]), (2.2, [2, 2]), (3.3, [3, 3, 3])], highlevel=False)
    assert isinstance(awkward1.to_awkward0(array), awkward0.Table)
    assert awkward1.to_awkward0(array).tolist() == [(0, []), (1.1, [1]), (2.2, [2, 2]), (3.3, [3, 3, 3])]
    assert isinstance(awkward1.to_awkward0(array[2]), tuple)
    assert awkward1.to_awkward0(array[2])[0] == 2.2
    assert awkward1.to_awkward0(array[2])[1].tolist() == [2, 2]

    array = awkward1.from_iter([0.0, [], 1.1, [1], 2.2, [2, 2], 3.3, [3, 3, 3]], highlevel=False)
    assert isinstance(awkward1.to_awkward0(array), awkward0.UnionArray)
    assert awkward1.to_awkward0(array).tolist() == [0.0, [], 1.1, [1], 2.2, [2, 2], 3.3, [3, 3, 3]]

    array = awkward1.from_iter([1.1, 2.2, None, None, 3.3, None, 4.4], highlevel=False)
    assert isinstance(awkward1.to_awkward0(array), awkward0.IndexedMaskedArray)
    assert awkward1.to_awkward0(array).tolist() == [1.1, 2.2, None, None, 3.3, None, 4.4]

    content = awkward1.layout.NumpyArray(numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]))
    index = awkward1.layout.Index64(numpy.array([3, 2, 2, 5, 0], dtype=numpy.int64))
    array = awkward1.layout.IndexedArray64(index, content)
    assert isinstance(awkward1.to_awkward0(array), awkward0.IndexedArray)
    assert awkward1.to_awkward0(array).tolist() == [3.3, 2.2, 2.2, 5.5, 0.0]

def test_fromawkward0():
    array = numpy.array([1.1, 2.2, 3.3, 4.4, 5.5])
    assert isinstance(awkward1.from_awkward0(array), awkward1.highlevel.Array)
    assert isinstance(awkward1.from_awkward0(array, highlevel=False), awkward1.layout.NumpyArray)
    assert awkward1.to_list(array) == [1.1, 2.2, 3.3, 4.4, 5.5]

    array = (123, numpy.array([1.1, 2.2, 3.3]))
    assert isinstance(awkward1.from_awkward0(array), awkward1.highlevel.Record)
    assert isinstance(awkward1.from_awkward0(array).layout, awkward1.layout.Record)
    assert awkward1.to_list(awkward1.from_awkward0(array)) == (123, [1.1, 2.2, 3.3])

    array = {"x": 123, "y": numpy.array([1.1, 2.2, 3.3])}
    assert isinstance(awkward1.from_awkward0(array), awkward1.highlevel.Record)
    assert isinstance(awkward1.from_awkward0(array).layout, awkward1.layout.Record)
    assert awkward1.to_list(awkward1.from_awkward0(array)) == {"x": 123, "y": [1.1, 2.2, 3.3]}

    array = awkward0.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
    assert isinstance(awkward1.from_awkward0(array, highlevel=False), (awkward1.layout.ListOffsetArray32, awkward1.layout.ListOffsetArrayU32, awkward1.layout.ListOffsetArray64))
    assert awkward1.to_list(awkward1.from_awkward0(array)) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]

    array = awkward0.fromiter([{"x": 0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": [2, 2]}])
    assert isinstance(awkward1.from_awkward0(array, highlevel=False), awkward1.layout.RecordArray)
    assert not awkward1.from_awkward0(array, highlevel=False).istuple
    assert awkward1.from_awkward0(array).layout.keys() == ["x", "y"]
    assert awkward1.to_list(awkward1.from_awkward0(array)) == [{"x": 0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": [2, 2]}]

    array = awkward0.Table([0.0, 1.1, 2.2], awkward0.fromiter([[], [1], [2, 2]]))
    assert isinstance(awkward1.from_awkward0(array, highlevel=False), awkward1.layout.RecordArray)
    assert awkward1.from_awkward0(array, highlevel=False).istuple
    assert awkward1.from_awkward0(array).layout.keys() == ["0", "1"]
    assert awkward1.to_list(awkward1.from_awkward0(array)) == [(0.0, []), (1.1, [1]), (2.2, [2, 2])]

    array = awkward0.fromiter([0.0, [], 1.1, [1], 2.2, [2, 2], 3.3, [3, 3, 3]])
    assert isinstance(awkward1.from_awkward0(array, highlevel=False), (awkward1.layout.UnionArray8_32, awkward1.layout.UnionArray8_U32, awkward1.layout.UnionArray8_64))
    assert awkward1.to_list(awkward1.from_awkward0(array)) == [0.0, [], 1.1, [1], 2.2, [2, 2], 3.3, [3, 3, 3]]

    array = awkward0.fromiter([1.1, 2.2, None, None, 3.3, None, 4.4])
    assert isinstance(awkward1.from_awkward0(array, highlevel=False), awkward1.layout.ByteMaskedArray)
    assert awkward1.to_list(awkward1.from_awkward0(array)) == [1.1, 2.2, None, None, 3.3, None, 4.4]

    array = awkward0.fromiter(["hello", "you", "guys"])
    assert isinstance(awkward1.from_awkward0(array, highlevel=False), (awkward1.layout.ListArray32, awkward1.layout.ListArrayU32, awkward1.layout.ListArray64, awkward1.layout.ListOffsetArray32, awkward1.layout.ListOffsetArrayU32, awkward1.layout.ListOffsetArray64))
    assert awkward1.from_awkward0(array, highlevel=False).parameters["__array__"] in ("string", "bytestring")
    assert awkward1.from_awkward0(array, highlevel=False).content.parameters["__array__"] in ("char", "byte")
    assert awkward1.to_list(awkward1.from_awkward0(array)) == ["hello", "you", "guys"]

    class Point(object):
        def __init__(self, x, y):
            self.x, self.y = x, y
        def __repr__(self):
            return "Point({0}, {1})".format(self.x, self.y)

    array = awkward0.fromiter([Point(1.1, 10), Point(2.2, 20), Point(3.3, 30)])
    assert awkward1.to_list(awkward1.from_awkward0(array)) == [{"x": 1.1, "y": 10}, {"x": 2.2, "y": 20}, {"x": 3.3, "y": 30}]
    assert "__record__" in awkward1.from_awkward0(array).layout.parameters

    array = awkward0.ChunkedArray([awkward0.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]]), awkward0.fromiter([[6.6]]), awkward0.fromiter([[7.7, 8.8], [9.9, 10.0, 11.1, 12.2]])])
    assert awkward1.to_list(awkward1.from_awkward0(array)) == [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8], [9.9, 10.0, 11.1, 12.2]]

    def generate1():
        return awkward0.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])

    def generate2():
        return awkward0.fromiter([[6.6]])

    def generate3():
        return awkward0.fromiter([[7.7, 8.8], [9.9, 10.0, 11.1, 12.2]])

    array = awkward0.ChunkedArray([awkward0.VirtualArray(generate1), awkward0.VirtualArray(generate2), awkward0.VirtualArray(generate3)])
    assert awkward1.to_list(awkward1.from_awkward0(array)) == [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8], [9.9, 10.0, 11.1, 12.2]]
